/***

Purpose
  Compute item and person fit statistics

Input data format

   1. Item Parameters: An Excel file containing the following labeled columns: 
                     'IID' (item ID, which must be from 1 to the number of items), 
                     'Difficulty' (item difficulty), 
                     'Threshold1~ThresholdN' (item thresholds), 
                     'K'(number of score categories).
   2. Ability Parameters: An Excel file containing the following labeled columns: 
                     'PID' (person ID, which must be from 1 to the number of persons) 
                     'Ability' (ability estimates).
   3. Score Data: A text file containing only the item responses in the following format: 
                     C1-CI: single digit numerically scored item responses, one column per scored response for each of the I items.

Input Macro variables
  Nitem: Number of items
  Nperson: Number of persons

Output files
   1. person_fit.txt: person fit statisics file, including 
                     'PID'(person ID), 
                     'MSU_person'(unweighted mean square),
                     'ZMSU_person'(standardized unweighted mean square),
                     'MSW_person'(weighted mean square), and
                     'ZMSW_person'(standardized weighted mean square).
   2. item_fit.txt: item fit statistics file, including 
                     'IID' (item ID),
                     'MSU_item'(unweighted mean square),
                     'ZMSU_item'(standardized unweighted mean square),
                     'MSW_item'(weighted mean square), and
                     'ZMSW_item'(standardized weighted mean square).
   3. expected_residual.txt: expected and residual score file, including
                     'PID'(person ID), 
                     'IID' (item ID),
                     'Expected' (expected score), and
                     'Residual' (residual score).

   4. residual_matrix.txt: matrix of residuals, including
                     C1-CI: 10 digit residual for each person-by-item combination in the original data file
                          In row n, there are I 10-digit values, one for each of the I items in the same sequence as the original data file.
                           
***/

*******************************************************
*** Enter file names and macro variable values here ***
*******************************************************;

%let path               =C:\Documents and Settings\Fit\Sample_input_output; /*Please specify the location of the input files*/
%let item_parameters    =Item_para.xls;
%let ability_parameters =Ability.xls;
%let score_data         =scoredata.TXT;
%let Nitem              =20;
%let Nperson            =100;

**************************
*** Macro begins here ***
**************************;

%macro fit;

**************************
*** Step 1: Input data ***
**************************;
*** File 1: estimated item parameters;
proc import out=item_para datafile="&path\&item_parameters."  dbms=excel replace;
     range="sheet1$"; getnames=yes; mixed=no; 
     scantext=yes; usedate=yes; scantime=yes;
run;

*** File 2: estimated ability;
proc import out=Ability datafile="&path\&ability_parameters."  dbms=excel replace;
     range="sheet1$"; getnames=yes; mixed=no; 
     scantext=yes; usedate=yes; scantime=yes;
run;

*** File 3: observed scores;
data Observed;
   infile "&path\&score_data." missover;
   input @1 (Item1-Item&Nitem) (1.);
   PID = _n_;
run;

*** Transpose the observed scores, create ID for items, for merging purpose;
proc transpose data=Observed out=Observed name=Item prefix=score;
   by PID;
   var Item1-Item&Nitem;
run;

data Observed;
set Observed;
IID=input(substr(Item,5,4), 8.);
run;

***************************************
*** Step 2: Compute expected scores ***
***************************************;
*** Maximum number of score categories;
proc means data=item_para noprint; 
   var K;
   output out=K_max max=K_max;
run;

data _null_; set K_max; call symput('K_max',K_max); run;

*** Expected scores;
proc sql;
   create table Expected as
   select a.*, b.*
     from Ability as a, item_para as b;
quit;

data Expected;
   set Expected;

   num0 = 1;
   %do i=1 %to (&K_max-1);
	   num&i = IFN(&i+1<=K,1,0)*exp(&i*ability-&i*difficulty%do j=1 %to &i;-coalesce(threshold&j,0) %end;) ;
   %end;
          
   denom = num0 %do i=1 %to (&K_max-1); +num&i %end; ;

   %do i=0 %to (&K_max-1); 
       prob&i = num&i/denom;
   %end;

   expected = 0*prob0 %do i=1 %to (&K_max-1); +&i*prob&i %end; ;

run;


**************************************
*** Step 3: Compute fit statistics ***
**************************************;
*** Residuals;
proc sql;
   create table resid as
   select a.*, b.score1 as observed, b.score1-a.expected as resid 
     from Expected as a, Observed as b
    where a.PID=b.PID and a.IID=b.IID;
quit;

data resid; 
    set resid;
    resid_var=(prob0*(0-expected)**2)%do i=1 %to (&K_max-1);+(prob&i*(&i-expected)**2)%end; ;
    resid_kur=(prob0*(0-expected)**4)%do i=1 %to (&K_max-1);+(prob&i*(&i-expected)**4)%end; ;

    if observed=. then do;
	resid_var=.;
	resid_kur=.;
	end; 

    zresid=resid/((resid_var)**.5);
    zresidsq=zresid**2;
    innerprod=zresidsq*resid_var;
	umsvar=resid_kur/(resid_var**2);
	wmsvar=resid_kur-(resid_var**2);
run;



***Compute person fit statistics;
proc sql;
    create table person_fit as
	select PID, mean(zresidsq) as MSU_person, 
                ((sum(umsvar)/count(observed)**2)-(1/count(observed)))**.5 as ums_sd_person,
	          sum(innerprod)/sum(resid_var) as MSW_person, 
			  (sum(wmsvar)/(sum(resid_var)**2))**.5 as wms_sd_person
	   from resid
   group by PID;        
quit;

data person_fit;
   set person_fit;
   ZMSU_person=((msu_person**(1/3))-1)*(3/ums_sd_person)+(ums_sd_person/3);
   ZMSW_person=((msw_person**(1/3))-1)*(3/wms_sd_person)+(wms_sd_person/3);
   drop ums_sd_person wms_sd_person;
run;

***Compute item fit statistics;
proc sql;
    create table item_fit as
	select IID, mean(zresidsq) as MSU_item, 
              ((sum(umsvar)/count(observed)**2)-(1/count(observed)))**.5 as ums_sd_item,
	          sum(innerprod)/sum(resid_var) as MSW_item, 
			  (sum(wmsvar)/(sum(resid_var)**2))**.5 as wms_sd_item
	   from resid
   group by IID;
quit;

data item_fit;
   set item_fit;
   ZMSU_item=((msu_item**(1/3))-1)*(3/ums_sd_item)+(ums_sd_item/3);
   ZMSW_item=((msw_item**(1/3))-1)*(3/wms_sd_item)+(wms_sd_item/3);
   drop ums_sd_item wms_sd_item;
run;


******************************
*** Step 4: Output Results ***
******************************;
data _null_;
   file "&path\person_fit.txt";
   if _n_=1 then do;
   put @1  'PID'
       @11 'MSU_person'
       @31 'ZMSU_person'
       @51 'MSW_person'
       @71 'ZMSW_person'; 
   end;

   set person_fit;
   put @1 PID @11 MSU_person @31 ZMSU_person @51 MSW_person @71 ZMSW_person ;
run;

data _null_;
   file "&path\item_fit.txt";
   if _n_=1 then do;
   put @1  'IID'
       @11 'MSU_item'
       @31 'ZMSU_item'
       @51 'MSW_item'
       @71 'ZMSW_item'; 
   end;

   set item_fit;
   put @1 IID @11 MSU_item @31 ZMSU_item @51 MSW_item @71 ZMSW_item ;
run;

proc sort data=resid out=sort;
   by pid;
run;

proc transpose data=sort out=trans;
   by pid;
   var resid;
run;

data trans;
   file "&path.\residual_matrix.txt" lrecl=1000 pad;
   set trans;
   put (col1-col&nitem.) (10.5);
run;

data _null_;
   file "&path\expected_residual.txt";
   if _n_=1 then do;
   put @1  'IID'
       @11 'PID'
       @31 'Expected'
       @51 'Residual';
   end;

   set Resid;
   put @1 IID @11 PID @31 Expected @51 Resid;
run;

%mend fit;
%fit;

  
     
